

LABEL_TRASH = -1
LABEL_NOISE = -2
LABEL_ALIEN = -9
LABEL_UNCLASSIFIED = -10
LABEL_NO_WAVEFORM = -11




to_name = { -1: 'Trash', 
    -2 : 'Noise',
    -9: 'Alien',
    -10: 'Unclassified',
    -11: 'No waveforms',
}

